--
-- Scriptfile for the project file generator
-- premake.sourceforge.net
--
-- This script is intended to work correctly for Visual Studio targets on
-- windows and the gnu target on linux and macosx,
-- don't use it with other targets.
--

-- ########################
-- ### Script variables ###
-- ########################
-- ### You may have to adjust these variables to fit your installation ###
-- ### To use static-crypto OPENSSL_LIB must be properly set           ###
if (windows) then
   OPENSSL_INCLUDE = ""
   OPENSSL_LIB = ""
   JDK_INCLUDE = ""
elseif (macosx) then
   --OPENSSL_INCLUDE = ""
   --OPENSSL_LIB = ""
   OPENSSL_INCLUDE = "/opt/local/include"
   OPENSSL_LIB = "/opt/local/lib"
   JDK_INCLUDE = "/System/Library/Frameworks/JavaVM.framework/Headers"
else
   OPENSSL_INCLUDE = "/usr/local/ssl/include"
   OPENSSL_LIB = "/usr/local/ssl/lib"
   JDK_INCLUDE = "/usr/lib/jvm/java-6-sun/include"
end



-- ########################
-- ### Project settings ###
-- ########################
-- Common settings
project.name = "aacskeys"
project.configs = { "Release", "Debug" }

-- Additional command line arguments
addoption("static-crypto", "Link static against OpenSSL (ignored on windows)")
addoption("no-ecdsa" , "Do not compile code that requires ECDSA")



-- #################################
-- ### Package aacskeys settings ###
-- #################################
-- Common settings
package = newpackage()
package.name = "aacskeys"
package.language = "c++"
package.kind = "exe"
package.files = { matchfiles("src/*.h", "src/*.cpp") }

-- Windows needs some additional sources to emulate POSIX
if (windows) then
   table.insert(package.files, matchfiles("src/win32/*.h", "src/win32/*.c"))
end

-- Outputdir dependend on os and configuration
if (windows) then
   package.config["Release"].bindir = "bin/win32"
   package.config["Debug"].bindir = "bin/win32/debug"
elseif (macosx) then
   package.config["Release"].bindir = "bin/macosx"
   package.config["Debug"].bindir = "bin/macosx/debug"
else
   package.config["Release"].bindir = "bin/linux"
   package.config["Debug"].bindir = "bin/linux/debug"
end

-- Wall for all configurations
package.buildflags = { "extra-warnings" }

-- Special buildflags for release builds
package.config["Release"].buildflags = { "no-symbols", "no-frame-pointer" }

-- Buildoptions dependend on compiler
if (windows) then
   table.insert(package.buildflags, "static-runtime")
   table.insert(package.config["Release"].buildflags, "optimize")
else
   package.config["Release"].buildoptions = { "-O3", "-pipe" }
   package.config["Debug"].buildoptions = { "-pipe" }
end

-- Common Defines
package.config["Debug"].defines = { "DEBUG" }

-- Defines dependend on compiler
if (windows) then
   -- Nothing for windows
else
   package.defines = { "_REENTRANT" }
end

-- Custom Defines
if (options["no-ecdsa"]) then
   table.insert(package.defines, "AACS_NO_ECDSA")
end

-- Common includes
if (string.len(OPENSSL_INCLUDE) > 0) then
   table.insert(package.includepaths, OPENSSL_INCLUDE)
end

-- Windows needs some additional includes for POSIX emulation
if (windows) then
   table.insert(package.includepaths, "src/win32")
end

-- Common libpaths
if ((string.len(OPENSSL_LIB) > 0) and (not options["static-crypto"] or windows)) then
   package.libpaths = { OPENSSL_LIB }
else
   package.libpaths = { }
end

-- Links dependend on os and configuration and custom
if (windows) then
   package.links = { "wsock32", "advapi32", "gdi32", "user32" }
   package.config["Release"].links = { "libeay32MT" }
   package.config["Debug"].links = { "libeay32MTd" }
else
   -- Custom links
   if (not options["static-crypto"]) then
      package.links = { "crypto" }
   else
      table.insert(package.linkoptions, OPENSSL_LIB .. "/libcrypto.a")
   end
   -- Dependend on os
   if (macosx) then
      table.insert(package.linkoptions, "-framework CoreFoundation -framework IOKit")
   end
end

-- Object directores dependend on os and configuration
if (windows) then
   package.config["Release"].objdir = "obj/win32/Release"
   package.config["Debug"].objdir = "obj/win32/Debug"
elseif (macosx) then
   package.config["Release"].objdir = "obj/macosx/Release"
   package.config["Debug"].objdir = "obj/macosx/Debug"
else
   package.config["Release"].objdir = "obj/linux/Release"
   package.config["Debug"].objdir = "obj/linux/Debug"
end



-- ####################################
-- ### Package libaacskeys settings ###
-- ####################################
-- Common settings
package = newpackage()
package.name = "libaacskeys"
package.target = "aacskeys"
package.language = "c++"
package.kind = "dll"
-- Because dylib can't be used the extension must be fixed manually
if (macosx) then
   package.targetextension = "dylib"
end
package.files = { matchfiles("src/*.h", "src/*.cpp") }

-- Windows needs some additional sources to emulate POSIX
if (windows) then
   table.insert(package.files, matchfiles("src/win32/*.h", "src/win32/*.c"))
end

-- Outputdir dependend on os and configuration
-- Note: Shared libraries will be put into bindir!
if (windows) then
   package.config["Release"].bindir = "lib/win32"
   package.config["Debug"].bindir = "lib/win32/debug"
   -- On windows import libraries get created, they will be put into libdir
   package.config["Release"].libdir = "lib/win32"
   package.config["Debug"].libdir = "lib/win32/debug"
elseif (macosx) then
   package.config["Release"].bindir = "lib/macosx"
   package.config["Debug"].bindir = "lib/macosx/debug"
else
   package.config["Release"].bindir = "lib/linux"
   package.config["Debug"].bindir = "lib/linux/debug"
end

-- Wall for all configurations
package.buildflags = { "extra-warnings" }

-- Special buildflags for release builds
package.config["Release"].buildflags = { "no-symbols", "no-frame-pointer" }

-- Buildoptions dependend on compiler
if (windows) then
   table.insert(package.buildflags, "static-runtime")
   table.insert(package.config["Release"].buildflags, "optimize")
else
   -- dylib can't be used because this also adds -flat_namespace which doesn't
   -- work on x86_64 with selfbuilt OpenSSL
   --if (macosx) then
   --   table.insert(package.buildflags, "dylib")
   --end
   package.config["Release"].buildoptions = { "-O3", "-pipe" }
   package.config["Debug"].buildoptions = { "-pipe" }
end

-- Common Defines
package.config["Debug"].defines = { "DEBUG" }

-- Defines dependend on compiler
if (windows) then
   package.defines = { "JNILIB" }
else
   package.defines = { "_REENTRANT", "JNILIB" }
end

-- Custom defines
if (options["no-ecdsa"]) then
   table.insert(package.defines, "AACS_NO_ECDSA")
end

-- Common includes
if (string.len(OPENSSL_INCLUDE) > 0) then
   table.insert(package.includepaths, OPENSSL_INCLUDE)
end
if (string.len(JDK_INCLUDE) > 0) then
   table.insert(package.includepaths, JDK_INCLUDE)
   -- Includes dependend on os
   if (windows) then
      table.insert(package.includepaths, JDK_INCLUDE .. "/win32")
   elseif (macosx) then
      -- No extra includes for macosx
   else
      table.insert(package.includepaths, JDK_INCLUDE .. "/linux")
   end
end

-- Windows needs some additional includes for POSIX emulation
if (windows) then
   table.insert(package.includepaths, "src/win32")
end

-- Common libpaths
if ((string.len(OPENSSL_LIB) > 0) and (not options["static-crypto"] or windows)) then
   package.libpaths = { OPENSSL_LIB }
else
   package.libpaths = { }
end

-- Links dependend on os and configuration and custom
if (windows) then
   package.links = { "wsock32", "advapi32", "gdi32", "user32" }
   package.config["Release"].links = { "libeay32MT" }
   package.config["Debug"].links = { "libeay32MTd" }
else
   -- Custom links
   if (not options["static-crypto"]) then
      package.links = { "crypto" }
   else
      table.insert(package.linkoptions, OPENSSL_LIB .. "/libcrypto.a")
   end
   -- Dependend on os
   if (macosx) then
      table.insert(package.linkoptions, "-dynamiclib, -framework CoreFoundation -framework IOKit")
   end
end

-- Object directores dependend on os and configuration
if (windows) then
   package.config["Release"].objdir = "obj/win32/ReleaseLib"
   package.config["Debug"].objdir = "obj/win32/DebugLib"
elseif (macosx) then
   package.config["Release"].objdir = "obj/macosx/ReleaseLib"
   package.config["Debug"].objdir = "obj/macosx/DebugLib"
else
   package.config["Release"].objdir = "obj/linux/ReleaseLib"
   package.config["Debug"].objdir = "obj/linux/DebugLib"
end
