/*
 * CMAC.cpp
 *
 * The One-key CBC MAC (CMAC, originally OMAC) message authentication code,
 * designed by T. Iwata and K. Kurosawa.
 *
 * @author Paulo S. L. M. Barreto
 *
 * This software is hereby placed in the public domain.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "cmac.h"

CMAC::CMAC(BlockCipher* E) {
    this->E = E;
    block_size = E->blockSize();
    if (block_size != 16 && block_size != 8) {
        throw "invalid block size";
    }
    block_ints = block_size >> 2;
    t = 0;
    mask = (block_size == 16) ? 0x87 : 0x1B;
    L = (uint *)calloc(block_ints, 4);
    V = (uint *)calloc(block_ints, 4);
    T = (byte *)calloc(block_size, 1);
    state = -1;
}

CMAC::~CMAC() {
    E = 0;
    block_size = 0;
    block_ints = 0;
    t = 0;
    mask = 0;
    memset(L, 0, block_size); free(L);
    memset(V, 0, block_size); free(V);
    memset(T, 0, block_size); free(T);
    state = -1;
}

void CMAC::init() {
    t = block_size;
    // compute padding mask:
    memset(L, 0, block_ints);
    E->encrypt(L, L); // L = E_K(0^n)
    uint c = (L[0] & 0x80000000U) != 0 ? mask : 0;
    for (uint b = 0; b < block_ints - 1; b++) {
        L[b] = (L[b] << 1) | (L[b + 1] >> 31);
    }
    L[block_ints - 1] = (L[block_ints - 1] << 1) ^ c; // B = 2L
    // initialize tag accumulator
    memset(T, (byte)0, block_size);
    state = 0; // initialized
}

void CMAC::update(const byte *M, uint m, bool aad) {
    if (state != 0) {
        throw "wrong order of operations";
    }
    byte *Z = &T[block_size - t];
    for (;;) {
        switch ((m > t) ? t : m) { // N.B. m is strictly larger than t!
        case 16: Z[15] ^= M[15]; // FALLTHROUGH
        case 15: Z[14] ^= M[14]; // FALLTHROUGH
        case 14: Z[13] ^= M[13]; // FALLTHROUGH
        case 13: Z[12] ^= M[12]; // FALLTHROUGH
        case 12: Z[11] ^= M[11]; // FALLTHROUGH
        case 11: Z[10] ^= M[10]; // FALLTHROUGH
        case 10: Z[ 9] ^= M[ 9]; // FALLTHROUGH
        case  9: Z[ 8] ^= M[ 8]; // FALLTHROUGH
        case  8: Z[ 7] ^= M[ 7]; // FALLTHROUGH
        case  7: Z[ 6] ^= M[ 6]; // FALLTHROUGH
        case  6: Z[ 5] ^= M[ 5]; // FALLTHROUGH
        case  5: Z[ 4] ^= M[ 4]; // FALLTHROUGH
        case  4: Z[ 3] ^= M[ 3]; // FALLTHROUGH
        case  3: Z[ 2] ^= M[ 2]; // FALLTHROUGH
        case  2: Z[ 1] ^= M[ 1]; // FALLTHROUGH
        case  1: Z[ 0] ^= M[ 0]; // FALLTHROUGH
        case  0: break;
        }
        if (m <= t) {
            break;
        }
        E->byte2int(T, V); E->encrypt(V, V); E->int2byte(V, T); // since there is more data, no padding applies
        // proceed to the next block:
        m -= t;
        M += t;
        t = block_size;
        Z = T;
        //assert(m > 0);
    }
    t -= m;
    //assert(m == 0 || t < block_size); // m == 0 here only occurs if m == 0 from the very beginning
}

void CMAC::update(const byte *M) {
    update(M, block_size, false);
}

void CMAC::finish(bool aad) {
    if (state == -1) {
        throw "wrong order of operations";
    } else if (state == 0) {
        // compute padding:
        if (t > 0) {
            // compute special padding mask:
            uint c = (L[0] & 0x80000000U) != 0 ? mask : 0;
            for (uint b = 0; b < block_ints - 1; b++) {
                L[b] = (L[b] << 1) | (L[b + 1] >> 31);
            }
            L[block_ints - 1] = (L[block_ints - 1] << 1) ^ c; // P = 4L
            // pad incomplete block:
            T[block_size - t] ^= 0x80; // padding toggle
            t = 0;
        }
        E->byte2int(T, V);
        for (uint b = 0; b < block_ints; b++) {
            V[b] ^= L[b];
        }
        E->encrypt(V, V); E->int2byte(V, T); // T contains the complete tag
        state = 1; // CMAC tag available
    }
}

void CMAC::getTag(byte *tag, uint t) {
    if (state == -1) {
        throw "wrong order of operations";
    } else if (state == 0) {
        finish(false);
    }
    memmove(tag, T, (t != 0 && t < block_size) ? t : block_size);
}

void CMAC::getTag(uint *tag) { // tag is one full block
    if (state == -1) {
        throw "wrong order of operations";
    } else if (state == 0) {
        finish(false);
    }
    E->byte2int(T, tag);
}
