/*
 * CMAC.h
 *
 * The One-key CBC MAC (CMAC, originally OMAC) message authentication code,
 * designed by T. Iwata and K. Kurosawa.
 *
 * @author Paulo S. L. M. Barreto
 *
 * This software is hereby placed in the public domain.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef __CMAC_H
#define __CMAC_H

#include "blockcipher.h"
#include "mac.h"

#ifndef USUAL_TYPES
#define USUAL_TYPES
typedef unsigned char   byte;
/* *** KenD00 start: conflicts under linux *** */
#if defined(WINDOWS) || defined(_WIN32) || defined(__APPLE__)
   #if defined(__LP64__)
      typedef unsigned int   uint;    /* assuming sizeof(uint) == 4 */
   #else
      typedef unsigned long   uint;   /* assuming sizeof(uint) == 4 */
   #endif
#endif
/* *** KenD00 end *** */
#endif /* USUAL_TYPES */

class CMAC: public MAC {

public:

    CMAC(BlockCipher* E);

    virtual ~CMAC();

    /**
     * Start computing a CMAC tag.
     */
    void init();

    /**
     * Update the CMAC tag computation with a message chunk.
     *
     * @param   M   message chunk
     * @param   m   its length in bytes
     */
    void update(const byte *M, uint m, bool aad = false);
    void update(const byte *M);

    /**
     * Complete the computation of the CMAC tag.
     */
    void finish(bool aad = false);

    /**
     * Get the complete MAC tag of the whole message provided.
     * @param   tag the MAC tag
     * @param   t   its length in bytes
     */
    void getTag(byte *tag, uint t);
    void getTag(uint *tag); // tag is one full block

    /**
     * Get the default tag size for the underlying block cipher.
     *
     * @return the default tag size in bytes.
     */
    uint tagSize() {
        return block_size;
    }

private:

    BlockCipher *E;    // block cipher context
    uint block_size;
    uint block_ints;
    uint t;             // remaining space on T, in bytes
    uint mask;
    int state;
    uint *L;            // CMAC padding: B = 2L, P = 4L
    uint *V;
    byte *T;            // CMAC tag
};

#endif /* __CMAC_H */
