#include "ioctl.h"

#include <cstring>
#include <sstream>
#include <iomanip>


// ###################
// ### Common code ###
// ###################

Drive::Drive() : m_handle(NULL), m_sk(0), m_asc(0), m_ascq(0)
#ifdef __APPLE__
               , m_plugIn(NULL), m_mmc(NULL), m_taskInterface(NULL), m_bluray(false)
#endif
{
   // Nothing more to do
}

Drive::~Drive() {
   close();
}

void Drive::sense_cmd(/*out*/ unsigned char *sk, /*out*/ unsigned char *asc, /*out*/ unsigned char *ascq) const {
   *sk = m_sk;
   *asc = m_asc;
   *ascq = m_ascq;
}

std::string Drive::sense_cmd() const {
   std::stringstream ss;
   ss << std::hex << std::uppercase << std::setfill('0');
   ss << "SK: 0x" << static_cast<unsigned int>(m_sk) << ", ASC: 0x" << std::setw(2) << static_cast<unsigned int>(m_asc) << ", ASCQ: 0x" << std::setw(2) << static_cast<unsigned int>(m_ascq);
   return ss.str();
}


#if defined(WINDOWS) || defined(_WIN32)
// ####################
// ### Windows side ###
// ####################

#include <winsock.h>
#include "my_ntddscsi.h"
#include "sam.h"


int Drive::open(const char *drive, bool exclusive) {
   // First close the drive to be sure its not open
   close();

   DRIVE_HANDLE h;
   UINT uDriveType;
   char szVolumeName[8];
   char szRootName[5];
   char drive_letter;

   drive_letter = drive[0];

   //Make drive_letter as a null terminated string in the form of e.g.: d:\.
   szRootName[0]=drive_letter;
   szRootName[1]=':';
   szRootName[2]='\\';
   szRootName[3]='\0';
   uDriveType = GetDriveType(szRootName);
   //#define GENERIC_READ 0x80000000L
   //#define GENERIC_WRITE 0x40000000L
   switch(uDriveType) {
       case DRIVE_CDROM: {
           //printf("Drive type is recognised as CDROM/DVD.\n");
           break;
       }
       default: {
           //printf("Drive type is not CDROM/DVD, aborting.\n");
           return -2;
       }
   }

   //Make drive_letter as a null terminated string in the form of e.g.: \\.\d:.
   szVolumeName[0]='\\';
   szVolumeName[1]='\\';
   szVolumeName[2]='.';
   szVolumeName[3]='\\';
   szVolumeName[4]=drive_letter;
   szVolumeName[5]=':';
   szVolumeName[6]='\0';
   //#define FILE_SHARE_READ 0x00000001
   //#define FILE_SHARE_WRITE 0x00000002
   //#define OPEN_EXISTING 3
   //#define FILE_ATTRIBUTE_NORMAL 0x00000080
   //This will only work for CD/DVD device on W2K and higher.
   h = CreateFile(szVolumeName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
   //#define INVALID_HANDLE_VALUE (long int *)-1
   if(h == INVALID_HANDLE_VALUE) {
       //Try again for NT4..
       h = CreateFile(szVolumeName, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
       if(h == INVALID_HANDLE_VALUE) {
          //printf("Could not create handle for CD/DVD device.\n");
          return -1;
       }
   }
   m_handle = h;
   return 0;
}

int Drive::close() {
   if (!m_handle) {
      return 0;
   }
   BOOL rv = CloseHandle(m_handle);
   m_handle = NULL;
   m_sk = 0;
   m_asc = 0;
   m_ascq = 0;
   if (rv) {
      return 0;
   } else {
      return -1;
   }
}

int Drive::send_cmd(const unsigned char *cmd, /*inout*/ unsigned char *buf, size_t send, size_t recv) {
   if (!m_handle) {
      return -4;
   }

   T_SPDT_SBUF sptd_sb;  //Includes sense buffer
   DWORD dwBytesReturned;

   sptd_sb.sptd.Length = sizeof(SCSI_PASS_THROUGH_DIRECT);
   sptd_sb.sptd.PathId = 0;    //SCSI card ID will be filled in automatically.
   sptd_sb.sptd.TargetId = 0;  //SCSI target ID will also be filled in.
   sptd_sb.sptd.Lun = 0;       //SCSI lun ID will also be filled in.
   sptd_sb.sptd.CdbLength = 12;  //CDB size.
   sptd_sb.sptd.SenseInfoLength = MAX_SENSE_LEN;  //Maximum length of sense data to retrieve.

   if (buf != NULL) {
      if (send > 0) {
         sptd_sb.sptd.DataIn = SCSI_IOCTL_DATA_OUT;  //There will be data going to the drive.
         sptd_sb.sptd.DataTransferLength = static_cast<ULONG>(send);
         sptd_sb.sptd.DataBuffer = static_cast<PVOID>(buf);
      } else if (recv > 0) {
         sptd_sb.sptd.DataIn = SCSI_IOCTL_DATA_IN;  //There will be data coming from the drive.
         sptd_sb.sptd.DataTransferLength = static_cast<ULONG>(recv);
         sptd_sb.sptd.DataBuffer = static_cast<PVOID>(buf);
      } else {
         sptd_sb.sptd.DataIn = SCSI_IOCTL_DATA_UNSPECIFIED;
         sptd_sb.sptd.DataTransferLength = 0;
         sptd_sb.sptd.DataBuffer = NULL;
      }
   } else {
      sptd_sb.sptd.DataIn = SCSI_IOCTL_DATA_UNSPECIFIED;
      sptd_sb.sptd.DataTransferLength = 0;
      sptd_sb.sptd.DataBuffer = NULL;
   }

   sptd_sb.sptd.TimeOutValue = 5;  //SCSI timeout value (max 5 sec).
   sptd_sb.sptd.SenseInfoOffset = sizeof(SCSI_PASS_THROUGH_DIRECT);

   //Copy CDB from cmd
   memcpy(sptd_sb.sptd.Cdb, cmd, 16);

   ZeroMemory(sptd_sb.SenseBuf, MAX_SENSE_LEN);

   //Send the command to drive.
   if (DeviceIoControl(m_handle, IOCTL_SCSI_PASS_THROUGH_DIRECT, static_cast<PVOID>(&sptd_sb), static_cast<DWORD>(sizeof(sptd_sb)), static_cast<PVOID>(&sptd_sb), static_cast<DWORD>(sizeof(sptd_sb)), &dwBytesReturned, NULL)) {
      // Store the sense information
      m_sk = sptd_sb.SenseBuf[2] & 0x0F;  //Sense key is only the lower 4 bits
      m_asc = sptd_sb.SenseBuf[12];
      m_ascq = sptd_sb.SenseBuf[13];
      // The command status MUST be checked because the I/O-Command can succeed but still return a status error!!!
      if (sptd_sb.sptd.ScsiStatus == STATUS_GOOD) {
         return 0;
      } else {
         return -2;
      }
   } else {
      return -1;
   }
}

#elif defined(__APPLE__)
// #####################
// ### Mac OS X side ###
// #####################

// For getmntinfo()
#include <sys/param.h>
#include <sys/ucred.h>
#include <sys/mount.h>

// For ioctl()
#include <fcntl.h>
#include <sys/ioctl.h>

// For opendev()/close()
#include <util.h>
#include <unistd.h>

// For BD-specific ioctl() constants
#include <IOKit/storage/IOBDMediaBSDClient.h>
#include <IOKit/storage/IODVDMediaBSDClient.h>
// For kIOBDMediaClass
#include <IOKit/storage/IOBDMedia.h>

// CoreFoundation
#include <CoreFoundation/CoreFoundation.h>
// IOKit
#include <IOKit/IOKitLib.h>
#include <IOKit/scsi-commands/SCSITaskLib.h>


// For Disk Arbitration
//#include <DiskArbitration/DiskArbitration.h>
//#include <DiskArbitration/DADisk.h>

// Get IO Media object from BSD Name
// http://cocoadev.com/forums/comments.php?DiscussionID=176
// http://developer.apple.com/samplecode/FSMegaInfo/listing7.html
// Left in for reference
// DiskArbitration can be used to unmount the volume without root rights,
// but it uses a callback and needs the current thread (CFRunLoop) to be halted
// until the callback has done its work. Right now exclusive access isn't needed
// that much that it satisfies this much work to do.
//int get_io_media_for_bsd_name(io_service_t *io_media, const char *bsd_name) {
//    DASessionRef session;
//    DADiskRef disk;
//
//    session = DASessionCreate(kCFAllocatorDefault);
//    if (!session) {
//        fprintf(stderr, "DASessionCreate() failed. \n");
//        return -1;
//    }
//
//    disk = DADiskCreateFromBSDName(kCFAllocatorDefault, session, bsd_name);
//    if (!disk) {
//        CFRelease(session);
//        fprintf(stderr, "DADiskCreateFromBSDName() failed. \n");
//        return -1;
//    }
//
//    *io_media = DADiskCopyIOMedia(disk);
//    if (!*io_media) {
//        CFRelease(session);
//        CFRelease(disk);
//        fprintf(stderr, "DADiskCopyIOMedia() failed. \n");
//        return -1;
//    }
//
//    CFRelease(session);
//    CFRelease(disk);
//
//    return 0;
//}


int Drive::getIOServiceForBSDName(/*out*/ io_service_t *io_service, const char *shortDevice) {
   kern_return_t ret;
   io_iterator_t iterator;

   // Get an iterator for the services matching the BSD Name
   // This will return only one entry, we use this method to get the error code
   ret = IOServiceGetMatchingServices(kIOMasterPortDefault, IOBSDNameMatching(kIOMasterPortDefault, 0, shortDevice), &iterator);
   if (ret != KERN_SUCCESS) {
      #ifdef DEBUG
      fprintf(stderr, "IOServiceGetMatchingServices: 0x%08x\n", ret);
      #endif
      return -1;
   }
   // Get the first entry and release the iterator
   *io_service = IOIteratorNext(iterator);
   IOObjectRelease(iterator);

   return 0;
}

int Drive::getMMCPlugIn(/*out*/ IOCFPlugInInterface ***MMCPlugIn, io_service_t io_service) {
   kern_return_t ret;
   IOCFPlugInInterface **plugInInterface = NULL;
   SInt32 score;

   bool firstRun = true;
   io_service_t parentService;

   #ifdef DEBUG
   io_name_t className;
   #endif

   // FIXME: This is an ugly fix to avoid to go up too high into the registry
   //        to get a plugin (like IDE-Controller, PCI-Bus or so), currently 5 steps
   //        reach up to the IOSCSIPeripheralDeviceNub, but this can change in the
   //        future so better find another way to stop trying to get a plugin
   for (int i = 0; i < 5; i++) {
      #ifdef DEBUG
      ret = IOObjectGetClass(io_service, className);
      if (ret != KERN_SUCCESS) {
         fprintf(stderr, "Failed to get class name: 0x%08X\n", ret);
      } else {
         fprintf(stderr, "Creating MMCPlugin for \"%s\"\n", className);
      }
      #endif

      ret = IOCreatePlugInInterfaceForService(io_service, kIOMMCDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score);
      if (ret == KERN_SUCCESS) {
         // Release the remaining intermediate io_service
         if (!firstRun) {
            IOObjectRelease(io_service);
         }
         *MMCPlugIn = plugInInterface;
         return 0;
      }

      #ifdef DEBUG
      fprintf(stderr, "Getting parent entry\n");
      #endif
      ret = IORegistryEntryGetParentEntry(io_service, kIOServicePlane, &parentService);
      if (ret == KERN_SUCCESS) {
         // Don't release the given first io_service, only the following intermediate ones
         if (firstRun) {
            firstRun = false;
         } else {
            IOObjectRelease(io_service);
         }
         io_service = parentService;
      } else {
         return -1;
      }
   }

   #ifdef DEBUG
   fprintf(stderr, "getMMCPlugIn loop finished\n");
   #endif
   return -2;
}

int Drive::getMMCInterface(/*out*/ MMCDeviceInterface ***MMCInterface, IOCFPlugInInterface **MMCPlugIn) {
   HRESULT herr;

   herr = (*MMCPlugIn)->QueryInterface(MMCPlugIn, CFUUIDGetUUIDBytes(kIOMMCDeviceInterfaceID), (LPVOID *)MMCInterface);
   if (herr != S_OK) {
      #ifdef DEBUG
      fprintf(stderr, "QueryInterface: 0x%X\n", herr);
      #endif
      return -1;
   }

   return 0;
}


int Drive::open(const char *drive, bool exclusive) {
   // First close the drive to be sure its not open
   close();

   // Return value, set by various failed operations, 0 == Success
   int rv = 0;

   // Pointer to the statfs structure that is currently processed
   struct statfs *fs = NULL;
   // Number of statfs structures
   int fs_count = 0;
   // Points to the found device, NULL if the device was not found
   char *device = NULL;
   // Short version of the device
   char *shortDevice = NULL;
   // Use this to specify the root drive (because drive itself is const)
   const char *rootDrive = "/";

   // If drive is empty, "rootify" it
   if (drive[0] == '\0') {
      drive = rootDrive;
   }
   // Determine the device from the volume
   fs_count = getmntinfo(&fs, 0);
   for (int i = 0; i < fs_count; i++) {
      if (!strcmp(fs[i].f_mntonname, drive)) {
         device = fs[i].f_mntfromname;
         break;
      }
   }
   if (device) {
      // Get the short device name (== without leading /dev)
      // Here we assume a path of the form /dev/<device>
      shortDevice = strrchr(device, '/');
      if (shortDevice) {
         // A '/' was found, the short path starts behind it
         shortDevice++;
      } else {
         // device seems to be already a short path
         shortDevice = device;
      }
   } else {
      // Not really necessary to set because we return at once
      rv = -2;
      return rv;
   }

   // IOService of the device
   io_service_t io_service;

   // First try to get the MMCDeviceInterface of the device
   // Get the IOService from the BSDName
   if (getIOServiceForBSDName(&io_service, shortDevice) == 0) {
      // Determine media type, poor mans method
      if (IOObjectConformsTo(io_service, kIOBDMediaClass)) {
         m_bluray = true;
      } else {
         m_bluray = false;
      }
      // Get the MMCDeviceInterfacePlugIn
      if (getMMCPlugIn(&m_plugIn, io_service) == 0) {
         // Get the MMCDeviceInterface
         if (getMMCInterface(&m_mmc, m_plugIn) != 0) {
            // Could not get the MMCDeviceInterface, destroy the PlugIn
            // and set the Plugin and Interface to NULL
            IODestroyPlugInInterface(m_plugIn);
            m_plugIn = NULL;
            // Just to be sure
            m_mmc = NULL;
            rv = -7;
            #ifdef DEBUG
            fprintf(stderr, "Failed to get MMCInterface\n");
            #endif
         }
      } else {
         // Just to be sure
         m_plugIn = NULL;
         rv = -6;
         #ifdef DEBUG
         fprintf(stderr, "Failed to get MMCPlugIn\n");
         #endif
      }
      // The IOService isn't used any more, release it
      IOObjectRelease(io_service);
   } else {
      rv = -5;
      #ifdef DEBUG
      fprintf(stderr, "Failed to get IOService for BSDName\n");
      #endif
   }

   if (exclusive) {
      // Exclusive access requires the MMCDeviceInterface
      if (m_mmc == NULL) {
         return rv;
      }
      // Get a SCSITaskDeviceInterface and obtain exclusive access
      m_taskInterface = (*m_mmc)->GetSCSITaskDeviceInterface(m_mmc);
      if (m_taskInterface) {
         if ((*m_taskInterface)->ObtainExclusiveAccess(m_taskInterface) != KERN_SUCCESS) {
            #ifdef DEBUG
            fprintf(stderr, "Failed to get exclusive access\n");
            #endif
            if (unmount(drive, 0) == 0) {
               if ((*m_taskInterface)->ObtainExclusiveAccess(m_taskInterface) != KERN_SUCCESS) {
                  #ifdef DEBUG
                  fprintf(stderr, "Failed to get exclusive access\n");
                  #endif
               } else {
                  return rv;
               }
            } else {
               #ifdef DEBUG
               fprintf(stderr, "Failed to unmount volume\n");
               #endif
            }
         } else {
            return rv;
         }
         (*m_taskInterface)->Release(m_taskInterface);
         m_taskInterface = NULL;
      } else {
         rv = -8;
         #ifdef DEBUG
         fprintf(stderr, "GetSCSITaskDeviceInterface returned NULL\n");
         #endif
      }
      // If we reach this code we failed to get exclusive access, clean up and return the error
      (*m_mmc)->Release(m_mmc);
      m_mmc = NULL;
      IODestroyPlugInInterface(m_plugIn);
      m_plugIn = NULL;
      if (rv == 0) {
         // Not really necessary, could return right here
         rv = -9;
      }
      return rv;
   } else {
      // Non-exclusive mode
      // Open the device
      // TODO: For some unknown reason open on the device does not work.
      //       I cannot open the block device, only the character device, why?
      m_handle = opendev(shortDevice, O_RDONLY | O_NONBLOCK, 0, NULL);
      if (m_handle < 0) {
         // Could not open the device, release possible used resources and return
         if (m_mmc) {
            (*m_mmc)->Release(m_mmc);
            m_mmc = NULL;
            IODestroyPlugInInterface(m_plugIn);
            m_plugIn = NULL;
         }
         m_handle = 0;
         // Not really necessary, could return right here
         rv = -1;
      } else {
         // We might got here with a missing MMCDeviceInterface
         rv = 0;
      }
      return rv;
   }
}

// Almost same as Linux
int Drive::close() {
   int rv = 0;
   if (!(m_handle || m_taskInterface)) {
      return rv;
   }

   // Release the SCSITaskInterface
   if (m_taskInterface) {
      (*m_taskInterface)->ReleaseExclusiveAccess(m_taskInterface);
      (*m_taskInterface)->Release(m_taskInterface);
      m_taskInterface = NULL;
   }
   // Release the MMCDeviceInterface
   if (m_mmc) {
      (*m_mmc)->Release(m_mmc);
      m_mmc = NULL;
   }
   // Destroy the PlugIn
   if (m_plugIn) {
      IODestroyPlugInInterface(m_plugIn);
      m_plugIn = NULL;
   }

   // Close the filehandle
   if (m_handle) {
      rv = ::close(m_handle);
      m_handle = 0;
   }

   // Reset the sense data
   m_sk = 0;
   m_asc = 0;
   m_ascq = 0;

   if (rv == 0) {
      return 0;
   } else {
      return -1;
   }
}

int Drive::send_cmd(const unsigned char *cmd, /*inout*/ unsigned char *buf, size_t send, size_t recv) {
   if (!(m_handle || m_taskInterface)) {
      return -4;
   }

   if (m_taskInterface) {
      // Exclusive mode
      SCSITaskInterface **task = NULL;
      SCSITaskStatus taskStatus = kSCSITaskStatus_No_Status;
      SCSI_Sense_Data senseData;
      SCSICommandDescriptorBlock cdb;
      UInt64 transferCount = 0;
      UInt8 dir;
      SCSITaskSGElement buffer;

      // Create the task
      task = (*m_taskInterface)->CreateSCSITask(m_taskInterface);
      if (!task) {
         return -1;
      }
      // Zero sense buffer and CDB
      memset(&senseData, 0, sizeof(senseData));
      memset(&cdb, 0, sizeof(cdb));

      // Set the CDB
      memcpy(&cdb, cmd, 16);
      if ((*task)->SetCommandDescriptorBlock(task, cdb, 16) != KERN_SUCCESS) {
         (*task)->Release(task);
         return -1;
      }
      // Set the transfer buffer
      // The definition of SCSITaskSGElement in SCSITaskLib.h differs under __LP64__
      // so we have to adjust the cast accordingly
      #if defined(__LP64__)
      buffer.address = (mach_vm_address_t)buf;
      #else
      buffer.address = (IOVirtualAddress)buf;
      #endif
      if (buf != NULL) {
         if (send > 0) {
            dir = kSCSIDataTransfer_FromInitiatorToTarget;
            buffer.length = send;
         } else if (recv > 0) {
            dir = kSCSIDataTransfer_FromTargetToInitiator;
            buffer.length = recv;
         } else {
            dir = kSCSIDataTransfer_NoDataTransfer;
            buffer.length = 0;
         }
      } else {
         dir = kSCSIDataTransfer_NoDataTransfer;
         buffer.length = 0;
      }
      if ((*task)->SetScatterGatherEntries(task, &buffer, 1, buffer.length, dir) != KERN_SUCCESS) {
         (*task)->Release(task);
         return -1;
      }
      // Set the timeout
      if ((*task)->SetTimeoutDuration(task, 5000) != KERN_SUCCESS) {
         (*task)->Release(task);
         return -1;
      }

      // Execute the Task
      if ((*task)->ExecuteTaskSync(task, &senseData, &taskStatus, &transferCount) != KERN_SUCCESS) {
         (*task)->Release(task);
         return -1;
      }
      (*task)->Release(task);
      m_sk = senseData.SENSE_KEY & kSENSE_KEY_Mask;
      m_asc = senseData.ADDITIONAL_SENSE_CODE;
      m_ascq = senseData.ADDITIONAL_SENSE_CODE_QUALIFIER;
      if (taskStatus == kSCSITaskStatus_GOOD) {
         return 0;
      } else {
         return -2;
      }
   } else {
      // Non-exclusive mode
      switch (cmd[0]) {
          // INQUIRY
          case 0x12: {
              if (m_mmc) {
                 SCSITaskStatus task_status = kSCSITaskStatus_No_Status;
                 SCSI_Sense_Data sense_data;
                 IOReturn io_return;

                 memset(&sense_data, 0, sizeof(sense_data));

                 SCSICmd_INQUIRY_StandardData inquiry_data;
                 memset(&inquiry_data, 0, sizeof(inquiry_data));

                 io_return = (*m_mmc)->Inquiry(m_mmc,
                     &inquiry_data, sizeof(inquiry_data),
                     &task_status, &sense_data);

                 if (io_return != kIOReturnSuccess) {
                     return -1;
                 }

                 if (sizeof(inquiry_data) < recv) {
                    memcpy(buf, &inquiry_data, sizeof(inquiry_data));
                 } else {
                    memcpy(buf, &inquiry_data, recv);
                 }
              } else {
                 // No MMCDeviceInterface, emulate with fake data
                 // This data can be retrieved from the IORegistry...
                 memset(buf, 0, recv);
                 if (recv >= 5) {
                    // Minimum size
                    buf[4] = 0x1F;
                    if (recv >= 12) {
                       // Vendor F0BA
                       buf[8] = 0x46;
                       buf[9] = 0x30;
                       buf[10] = 0x42;
                       buf[11] = 0x41;
                       if (recv >= 20) {
                          // Product F0BA
                          buf[16] = 0x46;
                          buf[17] = 0x30;
                          buf[18] = 0x42;
                          buf[19] = 0x41;
                          if (recv >= 36) {
                             // Revision F0BA
                             buf[32] = 0x46;
                             buf[33] = 0x30;
                             buf[34] = 0x42;
                             buf[35] = 0x41;
                          }
                       }
                    }
                 }
              }

              return 0;
          }
          // GET CONFIGURATION
          case 0x46: {
              if (m_mmc) {
                 SCSITaskStatus task_status = kSCSITaskStatus_No_Status;
                 SCSI_Sense_Data sense_data;
                 IOReturn io_return;

                 memset(&sense_data, 0, sizeof(sense_data));

                 io_return = (*m_mmc)->GetConfiguration(m_mmc,
                     cmd[1] & 0x02, // RT
                     ((cmd[2] & 0xFF) << 8) | (cmd[3] & 0xFF), // Feature Number
                     buf, recv,
                     &task_status, &sense_data);

                 if (io_return != kIOReturnSuccess) {
                     return -1;
                 }
              } else {
                 // No MMCDeviceInterface, emulate with fake data
                 memset(buf, 0, recv);
                 if (recv >= 8) {
                    // Feature header, report size for one Feature Descriptor
                    buf[3] = 0x0C;
                    // Set Profile to BD-ROM or HD-DVD ROM
                    if (m_bluray) {
                       buf[7] = 0x40;
                    } else {
                       buf[7] = 0x50;
                    }
                    // AACS Feature Descriptor
                    if (recv >= 10) {
                       // Size
                       buf[8] = 0x01;
                       buf[9] = 0x0D;
                       if (recv >= 11) {
                          // Mark Feature active
                          buf[10] = 0x01;
                          if (recv >= 12) {
                             // Fixed length
                             buf[11] = 0x04;
                             // No Binding Nonce Generation
                             // No Block Count
                             if (recv >= 15) {
                                // Only one AGID
                                buf[14] = 0x01;
                                // AACS Version 0 to indicate emulation
                             }
                          }
                       }
                    }
                 }
              }

              return 0;
          }
          // READ DISC STRUCTURE
          case 0xAD: {
             if (m_bluray) {
                dk_bd_read_structure_t read_disc_struct;
                memset(&read_disc_struct, 0, sizeof(read_disc_struct));

                read_disc_struct.format = cmd[7];
                read_disc_struct.address = cmd[2] << 24 | cmd[3] << 16 | cmd[4] << 8 | cmd[5];
                read_disc_struct.grantID = (cmd[10] >> 6) & 0x03;
                read_disc_struct.layer = cmd[6];

                read_disc_struct.buffer = buf;
                read_disc_struct.bufferLength = recv;

                if (ioctl(m_handle, DKIOCBDREADSTRUCTURE, &read_disc_struct) < 0) {
                   #ifdef DEBUG
                   perror("ioctl");
                   #endif
                   return -1;
                }
            } else {
               dk_dvd_read_structure_t read_disc_struct;
               memset(&read_disc_struct, 0, sizeof(read_disc_struct));

               read_disc_struct.format = cmd[7];
               read_disc_struct.address = cmd[2] << 24 | cmd[3] << 16 | cmd[4] << 8 | cmd[5];
               read_disc_struct.grantID = (cmd[10] >> 6) & 0x03;
               read_disc_struct.layer = cmd[6];

               read_disc_struct.buffer = buf;
               read_disc_struct.bufferLength = recv;

               if (ioctl(m_handle, DKIOCDVDREADSTRUCTURE, &read_disc_struct) < 0) {
                  #ifdef DEBUG
                  perror("ioctl");
                  #endif
                  return -1;
               }
            }

            return 0;
          }
          // REPORT KEY
          case 0xA4: {
             if (m_bluray) {
                dk_bd_report_key_t report_key;
                memset(&report_key, 0, sizeof(report_key));

                report_key.format = cmd[10] & 0x3F;
                report_key.keyClass = cmd[7];
                report_key.address = cmd[2] << 24 | cmd[3] << 16 | cmd[4] << 8 | cmd[5];
                report_key.grantID = (cmd[10] >> 6) & 0x3;

                report_key.bufferLength = recv;
                report_key.buffer = buf;

                if (ioctl(m_handle, DKIOCBDREPORTKEY, &report_key) < 0) {
                   #ifdef DEBUG
                   perror("ioctl");
                   #endif
                   return -1;
                }
             } else {
                dk_dvd_report_key_t report_key;
                memset(&report_key, 0, sizeof(report_key));

                report_key.format = cmd[10] & 0x3F;
                report_key.keyClass = cmd[7];
                report_key.address = cmd[2] << 24 | cmd[3] << 16 | cmd[4] << 8 | cmd[5];
                report_key.grantID = (cmd[10] >> 6) & 0x3;

                report_key.bufferLength = recv;
                report_key.buffer = buf;

                if (ioctl(m_handle, DKIOCDVDREPORTKEY, &report_key) < 0) {
                   #ifdef DEBUG
                   perror("ioctl");
                   #endif
                   return -1;
                }
             }

             return 0;
          }
          // SEND KEY
          case 0xA3: {
             if (m_bluray) {
                dk_bd_send_key_t send_key;
                memset(&send_key, 0, sizeof(send_key));

                send_key.format = cmd[10] & 0x3F;
                send_key.keyClass = cmd[7];
                send_key.grantID = (cmd[10] >> 6) & 0x3;

                send_key.bufferLength = send;
                send_key.buffer = buf;

                if (ioctl(m_handle, DKIOCBDSENDKEY, &send_key) < 0) {
                   #ifdef DEBUG
                   perror("ioctl");
                   #endif
                   return -1;
                }
             } else {
                dk_dvd_send_key_t send_key;
                memset(&send_key, 0, sizeof(send_key));

                send_key.format = cmd[10] & 0x3F;
                send_key.keyClass = cmd[7];
                send_key.grantID = (cmd[10] >> 6) & 0x3;

                send_key.bufferLength = send;
                send_key.buffer = buf;

                if (ioctl(m_handle, DKIOCDVDSENDKEY, &send_key) < 0) {
                   #ifdef DEBUG
                   perror("ioctl");
                   #endif
                   return -1;
                }
             }

             return 0;
          }
          default:
             // Not implemented CDB
             #ifdef DEBUG
             fprintf(stderr, "Unimplemented SCSI Command: 0x%02X\n", cmd[0]);
             #endif
             return -3;
      }
   }
}

#else
// ##################
// ### Linux side ###
// ##################

#include <linux/cdrom.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <mntent.h>


void Drive::cutTrailingSlashes(/*inout*/ char *path) {
   // Set the runner at the end of the string
   char *ptr = strrchr(path, '\0');

   // Go one step before the terminating 0
   ptr--;
   // Decrement the runner until we find the first non / or hit the beginning of the string
   while (ptr >= path && *ptr == '/') {
      // Cut off the path, if the path is root, the path will be empty
      *ptr = '\0';
      ptr--;
   }
}

void Drive::cutLastPathSegment(/*inout*/ char *path) {
   // This is duplicate code from cutTrailingSlashes but this way
   // we have to parse the path only once
   // Set the runner at the end of the string
   char *ptr = strrchr(path, '\0');

   // Go one step before the terminating 0
   ptr--;
   // Overjump trailing slashes
   while (ptr > path && *ptr == '/') {
      ptr--;
   }

   // Jump to the first non-slash
   while (ptr > path && *ptr != '/') {
      ptr--;
   }
   // Cut off the path, if the path is root, the path will be empty
   // TODO?: If the path contains only one non-root segment the path gets empty (==root!)
   *ptr = '\0';
}

int Drive::resolvePath(const char *path, /*out*/ char *target, size_t targetSize) {
   // To store stat information
   struct stat sourcestat;
   // Size of the work buffers
   size_t bufferSize = 8192;
   // Input buffer to read links into
   char inBuf[bufferSize];
   // Output buffer to construct absolute path
   char outBuf[bufferSize];
   // Current position in inBuf
   char *inPos = NULL;
   // Temporary pointer
   char *tmp = NULL;

   memset(inBuf, 0, bufferSize);
   memset(outBuf, 0, bufferSize);

   // Check if the system can resolve the path, if not we won't either
   if (stat(path, &sourcestat) != 0 && errno == ELOOP) {
      return -1;
   }

   // If the path is not absolute retrieve the cwd (this includes the empty path)
   if (path[0] != '/') {
      // Not absolute, get the cwd
      tmp = getcwd(outBuf, bufferSize);
      if (tmp == NULL) {
         return -2;
      }
      // To be safe
      cutTrailingSlashes(outBuf);
      #ifdef DEBUG
      fprintf(stderr, "Relative symlink, retrieved cwd: %s\n", outBuf);
      #endif
   }
   // Copy the path into the input buffer
   strncpy(inBuf, path, bufferSize - 1);

   // Resolve links until we have the final path
   while (true) {
      // First parse the path to translate all ./ and ../
      inPos = inBuf;
      if (*inPos == '/') {
         // Absolute path, clear the output
         outBuf[0] = '\0';
         inPos++;
         #ifdef DEBUG
         fprintf(stderr, "Absolute path, new output: %s, new input: %s\n", outBuf, inPos);
         #endif
      }
      while (*inPos != '\0') {
         if (*inPos == '/') {
            // Ignore multiple or trailing slashes
            inPos++;
            #ifdef DEBUG
            fprintf(stderr, "Ignoring /, new input: %s\n", inPos);
            #endif
            continue;
         }
         if (strncmp(inPos, ".", 1) == 0) {
            if (*(inPos + 1) == '/') {
               // Advance
               inPos += 2;
               #ifdef DEBUG
               fprintf(stderr, "Ignoring ./, new input: %s\n", inPos);
               #endif
               continue;
            } else if (*(inPos + 1) == '\0') {
               // Advance
               inPos += 1;
               #ifdef DEBUG
               fprintf(stderr, "Ignoring . at end of path, new input: %s\n", inPos);
               #endif
               continue;
            }
         }
         if (strncmp(inPos, "..", 2) == 0) {
            if (*(inPos + 2) == '/') {
               // Cut off last path segment of output
               cutLastPathSegment(outBuf);
               inPos += 3;
               #ifdef DEBUG
               fprintf(stderr, "Found ../, new output: %s, new input: %s\n", outBuf, inPos);
               #endif
               continue;
            } else if (*(inPos + 2) == '\0') {
               // Cut off last path segment of output
               cutLastPathSegment(outBuf);
               inPos += 2;
               #ifdef DEBUG
               fprintf(stderr, "Found .. at the end of path, new output: %s, new input: %s\n", outBuf, inPos);
               #endif
               continue;
            }
         }
         // Append first path segment to output
         char *tmp = strchr(inPos, '/');
         if (tmp != NULL) {
            *tmp = '\0';
            #ifdef DEBUG
            fprintf(stderr, "Path has more than one segments, current segment: %s\n", inPos);
            #endif
         }
         // First append a slash, the output path doesn't have a trailing slash
         strncat(outBuf, "/", bufferSize - strlen(outBuf) - 1);
         strncat(outBuf, inPos, bufferSize - strlen(outBuf) - 1);
         if (tmp != NULL) {
            inPos = tmp + 1;
         } else {
            inPos += strlen(inPos);
         }
         #ifdef DEBUG
         fprintf(stderr, "Copied path segment, new output: %s, new input: %s\n", outBuf, inPos);
         #endif
      }
      #ifdef DEBUG
      fprintf(stderr, "Processed output path: %s\n", outBuf);
      #endif

      // Check if we are dealing with a symlink
      // If the output path is empty we are at root, "rootify" it for lstat/readlink
      if (outBuf[0] == '\0') {
         strcpy(outBuf, "/");
      }
      if (lstat(outBuf, &sourcestat) != 0 || !S_ISLNK(sourcestat.st_mode)) {
         // Not a symlink, copy the current output and return
         strncpy(target, outBuf, targetSize - 1);
         #ifdef DEBUG
         fprintf(stderr, "Resolved target: %s\n", target);
         #endif
         return 0;
      }
      // Is a symlink, first read it
      ssize_t linkSize = readlink(outBuf, inBuf, bufferSize - 1);
      if (linkSize < 0) {
         return -3;
      }
      inBuf[linkSize] = '\0';
      // Now cut off the last path segment of the output path
      cutLastPathSegment(outBuf);
      #ifdef DEBUG
      fprintf(stderr, "Current link: %s, new output: %s\n", inBuf, outBuf);
      #endif
   }
}

int Drive::open(const char *drive, bool exclusive) {
   // First close the drive to be sure its not open
   close();

   // Filepointer for /proc/mounts
   FILE* mounts = NULL;
   // Pointer to the static struct from getmntent(FILE*)
   // endmntent(FILE*) does not free this struct (says the doc), anyway, can (should) static buffers be freed?
   struct mntent* mountEntry = NULL;
   // Size of buffer for resolving the absolute drive path
   size_t bufferSize = 8192;
   // Buffer for for resolving the absolute drive path
   char buffer[bufferSize];

   if (resolvePath(drive, buffer, bufferSize) < 0) {
      return -4;
   }

   // Open /proc/mounts for parsing
   if ((mounts = setmntent("/proc/mounts", "r")) != NULL) {
      // Get mount entries until NULL is returned
      while ((mountEntry = getmntent(mounts)) != NULL) {
         // Check if this entry contains our mountpoint
         // No need to escape characters, getmntent has taken care of it
         if (strcmp(mountEntry->mnt_dir, buffer) == 0) {
            // Its our mountpoint, close the file and open the device
            endmntent(mounts);
            int rv = ::open(mountEntry->mnt_fsname, O_RDONLY | O_NONBLOCK);
            if (rv > -1) {
               m_handle = rv;
               return 0;
            } else {
               return -1;
            }
         }
      }
      // Reached EOF or an error occured, doesn't matter, our entry was not found
      endmntent(mounts);
      return -2;
   } else {
      // Could not open /proc/mounts
      return -3;
   }
}

int Drive::close() {
   if (!m_handle) {
      return 0;
   }
   int rv = ::close(m_handle);
   m_handle = 0;
   m_sk = 0;
   m_asc = 0;
   m_ascq = 0;
   if (rv == 0) {
      return 0;
   } else {
      return -1;
   }
}

int Drive::send_cmd(const unsigned char *cmd, /*inout*/ unsigned char *buf, size_t send, size_t recv) {
   if (!m_handle) {
      return -4;
   }

   struct cdrom_generic_command cgc;
   struct request_sense sense;

   memset(&cgc, 0, sizeof(cgc));
   memcpy(cgc.cmd, cmd, CDROM_PACKET_SIZE);
   cgc.sense = &sense;
   cgc.timeout = 5000; // 5 secs timeout

   if (buf != NULL) {
      if (send > 0) {
         cgc.data_direction = CGC_DATA_WRITE;
         cgc.buflen = send;
         cgc.buffer = buf;
      } else if (recv > 0) {
         cgc.data_direction = CGC_DATA_READ;
         cgc.buflen = recv;
         cgc.buffer = buf;
      } else {
         cgc.data_direction = CGC_DATA_NONE;
         cgc.buflen = 0;
         cgc.buffer = NULL;
      }
   } else {
      cgc.data_direction = CGC_DATA_NONE;
      cgc.buflen = 0;
      cgc.buffer = NULL;
   }

   int rv = ioctl(m_handle, CDROM_SEND_PACKET, &cgc);
   // In contrast to the windows side when the MMC command fails
   // ioctl returns -1 (its usual error code) and not 0 with a
   // status != STATUS_GOOD (whats cgc.stat for then?)
   // So store the sense information and return -2 to signal
   // a MMC error.
   // What happens if a real I/O error happens, what
   // does the sense information contain then? How can both error
   // situations be distinguished? With errno?
   // Store sense information
   m_sk = sense.sense_key & 0x0F;
   m_asc = sense.asc;
   m_ascq = sense.ascq;
   if (rv < 0) {
      return -2;
   } else {
      return 0;
   }
}

#endif
