#ifndef IOCTL_H
#define IOCTL_H

#include <cstddef>  // For size_t
#include <string>
#if defined(WINDOWS) || defined(_WIN32)
   #include <windows.h>
   #define DRIVE_HANDLE HANDLE
#else
   #define DRIVE_HANDLE int
#endif

#if defined(__APPLE__)
   // Forward declarations
   #include <IOKit/IOTypes.h>
   #include <IOKit/IOCFPlugIn.h>
   //class io_service_t;
   //class IOCFPlugInInterface;
   class MMCDeviceInterface;
   class SCSITaskDeviceInterface;
#endif

/**
 * This class provides abstract access to an underlying physical drive to send CDB's to it.
 * It works under win32, linux and macosx and provides the same interface for both systems.
 *
 * TODO: The linux and macosx code do interpret a 0 FD as invalid but accept it as
 *       return value of open!
 */
class Drive {
private:
   /// The handle / filepointer of the open drive, NULL if no drive is opened
   DRIVE_HANDLE m_handle;
   /// Sense SK values of the last successful command
   unsigned char m_sk;
   /// Sense ASC value of the last successful command
   unsigned char m_asc;
   /// Sense ASCQ value of the last successful command
   unsigned char m_ascq;
#if defined(__APPLE__)
   /// User space plugin to communicate with kernel side
   IOCFPlugInInterface **m_plugIn;
   /// Used to send MMC commands to the drive, does not require exclusive access
   MMCDeviceInterface **m_mmc;
   /// Used to create SCSITasks in exclusive mode
   SCSITaskDeviceInterface **m_taskInterface;
   /*
    * True if the drive is in Blu-Ray mode, otherwise false
    * The filter code for non exclusive access mode requires the media type,
    * on macosx this can be derived from the IOMedia object
    */
   bool m_bluray;
#endif


public:
   Drive();
   ~Drive();

   /**
    * Opens the given drive.
    *
    * @param drive This is the platform specific path to the drive, e.g. under windows this is the drive letter, under linux the mountpoint
    * @param exclusive If true, use exclusive IO to access the drive. Only used under Mac OS X, ignored otherwise.
    * @return  0 == Success
    *         -1 == Error opening the drive
    *         -2 == Invalid drive type (windows), could not find device from mountpoint (linux)
    *         -3 == Could not open /proc/mounts (linux)
    *         -4 == Could not read symlink (linux)
    *         -5 == Could not get IOService (macosx)
    *         -6 == Could not get MMCPlugIn (macosx)
    *         -7 == Could not get MMCInterface (macosx)
    *         -8 == Could not get SCSITaskDeviceInterface (macosx)
    *         -9 == Could not get exclusive access (macosx)
    */
   int open(const char *drive, bool exclusive = false);
   /**
    * Closes the drive.
    *
    * @return  0 == Success
    *         -1 == Error closing the drive
    */
   int close();
   /**
    * Sends a CDB to the drive.
    * If send > 0 this amount of data is send from buf to the drive, if recv > 0 at most this amount of
    * data is stored into buf from the drive. Send and recv > 0 is not allowed (send is used in that case).
    *
    * On Mac OS X sense information is only available if using exclusive IO.
    *
    * @param cmd The CDB, MUST be 16 bytes long
    * @param buf Buffer used to send / receive data from the drive (MUST be at least as large as specified by send / receive
    * @param send If > 0 this amount of data from buf is send to the drive
    * @param recv If > 0 this amount of data is expected from the drive and stored into buf
    * @return  0 == Success
    *         -1 == I/O Error
    *         -2 == The CDB command failed
    *         -3 == The CDB command is not implemented (macosx)
    *         -4 == Drive not open
    */
   int send_cmd(const unsigned char *cmd, /*inout*/ unsigned char *buf = NULL, size_t send = 0, size_t recv = 0);
   /**
    * Returns the sense information from the last successfully send command (that means it's return value was 0 or -2).
    *
    * @param sk SENSE_KEY
    * @param asc ASC
    * @param ascq ASCQ
    */
   void sense_cmd(/*out*/ unsigned char *sk, /*out*/ unsigned char *asc, /*out*/ unsigned char *ascq) const;
   /**
    * Returns the sense information from the last successfully send command as formatted string (that means it's return value was 0 or -2).
    *
    * @return String of the form SK: <sk>, ASC: <asc>, ASCQ: <ascq>
    */
   std::string sense_cmd() const;

private:
#if defined(__APPLE__)
   /**
    * Gets the IOService for the given BSD device.
    *
    * @param io_service The IOService for the given BSD device
    * @param shortDevice The BSD device, has to be the short name without /dev
    * @return 0 if successful, -1 otherwise
    */
   int getIOServiceForBSDName(/*out*/ io_service_t *io_service, const char *shortDevice);
   /**
    * Gets the MMCDevicePlugIn for the given IOService.
    *
    * @param MMCPlugIn The MMCDevicePlugIn for the given IOService
    * @param io_service The IOService to use, navigates higher up the registry tree if not
    *                   successful with this one
    * @return 0 if successful, -1 otherwise
    */
   int getMMCPlugIn(/*out*/ IOCFPlugInInterface ***MMCPlugIn, io_service_t io_service);
   /**
    * Gets the MMCDeviceInterface for the given MMCDevicePlugIn.
    *
    * @param MMCInterface The MMCDeviceInterface for the given MMCDevicePlugIn
    * @param MMCPlugIn The MMCDevicePlugIn to use
    * @return 0 if successful, -1 otherwise
    */
   int getMMCInterface(/*out*/ MMCDeviceInterface ***MMCInterface, IOCFPlugInInterface **MMCPlugIn);
#endif
#if !defined(WINDOWS) && !defined(_WIN32) && !defined(__APPLE__)
   /**
    * Cuts off trailing slashes from the given path.
    *
    * If the path is root it becomes the empty string.
    *
    * @param path The path to process
    */
   void cutTrailingSlashes(/*inout*/ char *path);
   /**
    * Cuts off the last segment of the given path.
    *
    * If the path is root or the last path segment, the result will be the empty string.
    *
    * @param path The path to process.
    */
   void cutLastPathSegment(/*inout*/ char *path);
   /**
    * Resolves the given path, makes it absolute.
    * 
    * Follows all symlinks, ./ and ../ in the path until an absolute path remains.
    * The empty string resolves into the current working directory.
    * 
    * @param path The path to resolve
    * @param target The out buffer to write the resolved path into
    * @param targetSize Size of target
    * @return  0: Success
    *         -1: Too many symbolic links encountered while traversing the path
    *         -2: Error retrieving current working directory
    *         -3: Error reading symlink
    */
   int resolvePath(const char *path, /*out*/ char *target, size_t targetSize);
#endif
};

#endif
