#include "mmc.h"

#include <cstring>

#define CDB_SIZE 16


MMC::MMC(Drive *drive, MediaType mt) : m_drive(drive), m_mt(mt) {
   // Nothing more to do
}

MMC::~MMC() {
   // Nothing to do
}


MMC::MediaType MMC::getMediaType() const {
   return m_mt;
}

void MMC::setMediaType(MMC::MediaType mt) {
   m_mt = mt;
}

// These are the MMC commands (some may be composed of multiple commands actually), they return the same values as Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t)
int MMC::read_drive_info(/*out*/ char *drive_info) {
   unsigned char cmd[CDB_SIZE];
   memset(cmd, 0, CDB_SIZE);
   unsigned char buf[96];
   memset(buf, 0, 96);

   // INQUIRY
   cmd[0] = 0x12;
   // 96 bytes allocation length
   // (big enough to return all data except vendor specific parameters)
   cmd[4] = 0x60;

   int rv = m_drive->send_cmd(cmd, buf, 0, 96);
   if (rv == 0) {
      // Check if the data is at least as big to store PRODUCT REVISION LEVEL
      if (buf[4] < 31) {
         // Not big enough, this is not allowed, but don't return an error, just signal it
         strcpy(drive_info, "#UNKNOWN#");
      } else {
         // PRODUCT REVISION LEVEL + Vendor Specific can have at most this count of characters
         // because of the reported ADDITIONAL LENGTH field
         unsigned char maxChars = buf[4] - 27;
         // Limit the chars to the maximum size of PRODUCT REVISION LEVEL + Vendor Specific
         if (maxChars > 24) {
            maxChars = 24;
         }
         strncpy(drive_info, reinterpret_cast<char*>(buf + 32), maxChars);
         // To ensure that the string is 0 terminated
         drive_info[maxChars] = '\0';
      }
   }
   return rv;
}

int MMC::read_aacs_feature(/*out*/ unsigned char *version, /*out*/ unsigned char *concurrent_agids, /*out*/ bool *bng, /*out*/ unsigned char *block_count, /*out*/ bool *enabled) {
   unsigned char cmd[CDB_SIZE];
   memset(cmd, 0, CDB_SIZE);
   // The AACS Feature Descriptor and the header takes 16 bytes
   unsigned char buf[16];
   memset(buf, 0, 16);

   // GET CONFIGURATION
   cmd[0] = 0x46;
   // Request Feature Header + Descriptor from Starting Number
   cmd[1] = 0x02;
   // AACS Feature Descriptor Number 0x010D
   cmd[2] = 0x01;
   cmd[3] = 0x0D;
   // Our buffer is 16 bytes
   cmd[7] = 0x00;
   cmd[8] = 0x10;

   int rv = m_drive->send_cmd(cmd, buf, 0, 16);
   if (rv == 0) {
      unsigned int data_length = buf[0];
      data_length <<= 8;
      data_length |= buf[1];
      data_length <<= 8;
      data_length |= buf[2];
      data_length <<= 8;
      data_length |= buf[3];
      // The remaining Feature Header is 4 bytes + the 8 bytes of the AACS Feature Descriptor
      if (data_length >= 12) {
         if (version != NULL) {
            *version = buf[15];
         }
         if (concurrent_agids != NULL) {
            *concurrent_agids = buf[14] & 0x0F;
         }
         if (bng != NULL) {
            *bng = buf[12] & 0x01;
         }
         if (block_count != NULL) {
            *block_count = buf[13];
         }
         if (enabled != NULL) {
            *enabled = buf[10] & 0x01;
         }
      } else {
         if (version != NULL) {
            *version = 0;
         }
         if (concurrent_agids != NULL) {
            *concurrent_agids = 0;
         }
         if (bng != NULL) {
            *bng = false;
         }
         if (block_count != NULL) {
            *block_count = 0;
         }
         if (enabled != NULL) {
            *enabled = false;
         }
      }
   }
   return rv;
}

int MMC::report_agid(/*out*/ unsigned char *agid) {
   unsigned char buf[8];
   memset(buf, 0, 8);

   int rv = report_key(0, 0, 0, 0, buf, 8);
   if (rv == 0) {
      *agid = (buf[7] & 0xFF) >> 6;
   }
   return rv;
}

int MMC::invalidate_agid(unsigned char agid) {
   // The ioctl command under macosx doesn't like a NULL buffer so supply a
   // dummy buffer big enough to store the data length
   unsigned char buf[2];
   memset(buf, 0, 2);

   return report_key(agid, 0, 0, 0x3F, buf, 2);
}

int MMC::send_host_cert_chal(unsigned char agid, const unsigned char *nonce, const unsigned char *cert) {
   unsigned char buf[116];
   memset(buf, 0, 116);

   buf[1] = 0x72;
   memcpy(buf + 4, nonce, 20);
   memcpy(buf + 24, cert, 92);

   return send_key(agid, 0x01, buf, 116);
}

int MMC::send_host_key(unsigned char agid, const unsigned char *point, const unsigned char *signature) {
   unsigned char buf[84];
   memset(buf, 0, 84);

   buf[1] = 0x52;
   memcpy(buf + 4, point, 40);
   memcpy(buf + 44, signature, 40);
   return send_key(agid, 0x02, buf, 84);
}

int MMC::report_drive_cert_chal(unsigned char agid, /*out*/ unsigned char *nonce, /*out*/ unsigned char *cert) {
   unsigned char buf[116];
   memset(buf, 0, 116);

   int rv = report_key(agid, 0, 0, 0x01, buf, 116);
   if (rv == 0) {
      memcpy(nonce, buf + 4, 20);
      memcpy(cert, buf + 24, 92);
   }
   return rv;
}

int MMC::report_drive_key(unsigned char agid, /*out*/ unsigned char *point, /*out*/ unsigned char *signature) {
   unsigned char buf[84];
   memset(buf, 0, 84);

   int rv = report_key(agid, 0, 0, 0x02, buf, 84);
   if (rv == 0) {
      memcpy(point, buf + 4, 40);
      memcpy(signature, buf + 44, 40);
   }
   return rv;
}

int MMC::read_vid(unsigned char agid, /*out*/ unsigned char *vid, /*out*/ unsigned char *mac) {
   unsigned char cmd[CDB_SIZE];
   memset(cmd, 0, CDB_SIZE);
   unsigned char buf[36];
   memset(buf, 0, 36);

   // READ DISC STRUCTURE
   cmd[0] = 0xAD;
   cmd[1] = static_cast<unsigned char>(m_mt);
   // Volume Identifier of AACS
   cmd[7] = 0x80;
   cmd[8] = 0;
   // Enough space for the Dara Length field and the actual data
   cmd[9] = 0x24;
   cmd[10] = (agid << 6) & 0xC0;

   int rv = m_drive->send_cmd(cmd, buf, 0, 36);
   if (rv == 0) {
      memcpy(vid, buf + 4, 16);
      memcpy(mac, buf + 20, 16);
   }
   return rv;
}

int MMC::read_vid_mc08(unsigned char agid, /*out*/ unsigned char *vid, /*out*/ unsigned char *mac) {
   unsigned char cmd[CDB_SIZE];
   unsigned char buf[8];
   int rv = 0;

   //FIXME?: DF doesn't get disabled if any command after its enabling fails
   // First enable DF command
   memset(cmd, 0, CDB_SIZE);
   cmd[0] = 0x1D;
   cmd[4] = 0x08;
   memset(buf, 0, 8);
   buf[0] = 0x88;
   buf[3] = 0x04;
   buf[4] = 0x02;
   buf[5] = 0x6F;
   buf[6] = 0x01;
   buf[7] = 0x00;
   rv = m_drive->send_cmd(cmd, buf, 8);
   if (rv != 0) {
      return rv;
   }

   // Memset 60C1CC to 0000
   memset(cmd, 0, CDB_SIZE);
   cmd[0] = 0xDF;
   cmd[2] = 0xE3;
   // from address
   cmd[4] = 0x60;
   cmd[5] = 0xC1; 
   cmd[6] = 0xCC; 
   // to address
   cmd[7] = 0x60; 
   cmd[8] = 0xC1;
   cmd[9] = 0xCD; 
   // 0000
   cmd[10] = 0x00;
   cmd[11] = 0x00;
   rv = m_drive->send_cmd(cmd);
   if (rv != 0) {
      return rv;
   }

   // Memset 60C1CE to 0001
   memset(cmd, 0, CDB_SIZE);
   cmd[0] = 0xDF;
   cmd[2] = 0xE3;
   // from address
   cmd[4] = 0x60;
   cmd[5] = 0xC1; 
   cmd[6] = 0xCE; 
   // to address
   cmd[7] = 0x60;
   cmd[8] = 0xC1;
   cmd[9] = 0xCF; 
   // 0001
   cmd[10] = 0x00;
   cmd[11] = 0x01;
   rv = m_drive->send_cmd(cmd);
   if (rv != 0) {
      return rv;
   }

   // Memset 60C1C8 to 0000
   memset(cmd, 0, CDB_SIZE);
   cmd[0] = 0xDF;
   cmd[2] = 0xE3;
   // from address
   cmd[4] = 0x60;
   cmd[5] = 0xC1; 
   cmd[6] = 0xC8; 
   // to address
   cmd[7] = 0x60;
   cmd[8] = 0xC1;
   cmd[9] = 0xC9; 
   // 0000
   cmd[10] = 0x00;
   cmd[11] = 0x00;
   rv = m_drive->send_cmd(cmd);
   if (rv != 0) {
      return rv;
   }

   // Memset 60C1CA to 0005
   memset(cmd, 0, CDB_SIZE);
   cmd[0] = 0xDF;
   cmd[2] = 0xE3;
   // from address
   cmd[4] = 0x60;
   cmd[5] = 0xC1; 
   cmd[6] = 0xCA; 
   // to address
   cmd[7] = 0x60;
   cmd[8] = 0xC1;
   cmd[9] = 0xCB; 
   // 0005
   cmd[10] = 0x00;
   cmd[11] = 0x05;
   rv = m_drive->send_cmd(cmd);
   if (rv != 0) {
      return rv;
   }

   // Read the vid the normal way
   rv = read_vid(agid, vid, mac);
   if (rv != 0) {
      return rv;
   }

   // Lastly disable the DF command
   memset(cmd, 0, CDB_SIZE);
   cmd[0] = 0x1D;
   cmd[4] = 0x08;
   memset(buf, 0, 8);
   buf[0] = 0x88;
   buf[3] = 0x04;
   buf[4] = 0x02;
   buf[5] = 0x6F;
   buf[6] = 0x00;
   buf[7] = 0x00;
   rv = m_drive->send_cmd(cmd, buf, 8);

   return rv;
}

int MMC::read_bn(unsigned char agid, unsigned int lba, unsigned char block_count, /*out*/ unsigned char *bn, /*out*/ unsigned char *mac) {
   unsigned char buf[36];
   memset(buf, 0, 36);

   int rv = report_key(agid, lba, block_count, 0x21, buf, 36);
   if (rv == 0) {
      memcpy(bn, buf + 4, 16);
      memcpy(mac, buf + 20, 16);
   }
   return rv;
}


int MMC::report_key(unsigned char agid, unsigned int address, unsigned char block_count, unsigned char key_format, /*out*/ unsigned char *buffer, unsigned short length) {
   unsigned char cmd[CDB_SIZE];
   memset(cmd, 0, CDB_SIZE);
   if (buffer != NULL) {
      memset(buffer, 0, length);
   }

   // REPORT KEY
   cmd[0] = 0xA4;
   // Address / Reserved
   cmd[2] = (address >> 24) & 0xFF;
   cmd[3] = (address >> 16) & 0xFF;
   cmd[4] = (address >> 8) & 0xFF;
   cmd[5] = address & 0xFF;
   // Block Count / Reserved
   cmd[6] = block_count;
   // Key Class
   cmd[7] = 0x02;
   // Allocation Length
   cmd[8] = (length >> 8) & 0xFF;
   cmd[9] = length & 0xFF;
   // AGID and Key Format
   cmd[10] = (agid << 6) | (key_format & 0x3F);

   return m_drive->send_cmd(cmd, buffer, 0, length);
}

int MMC::send_key(unsigned char agid, unsigned char key_format, unsigned char *buffer, unsigned short length) {
   unsigned char cmd[CDB_SIZE];
   memset(cmd, 0, CDB_SIZE);

   // SEND KEY
   cmd[0] = 0xA3;
   // Key Class
   cmd[7] = 0x02;
   // Allocation Length
   cmd[8] = (length >> 8) & 0xFF;
   cmd[9] = length & 0xFF;
   // AGID and Key Format
   cmd[10] = (agid << 6) | (key_format & 0x3F);

   return m_drive->send_cmd(cmd, buffer, length);
}
