#ifndef MMC_H
#define MCC_H

#include "ioctl.h"


/**
 * This class is used to send MMC commands to a drive.
 */
class MMC {
public:
   /// Possible media types to use
   enum MediaType { HDDVD = 0, BLURAY = 1 };

private:
   /// The used Drive
   Drive *m_drive;
   /// The media type to use
   MediaType m_mt;

public:
   /**
    * Constructs a new MMC object whichs uses the given Drive.
    *
    * @param The drive to use. This class does not open or close the drive, this has to be done from outside
    */
   MMC(Drive *drive, MediaType mt = HDDVD);
   ~MMC();

   /**
    * Returns the current media type this class uses for the commands (where necessary).
    * Attention: This method DOES NOT query the media in the drive, it just returns the value set with setMediaType(MediaType)
    *
    * @return The current set media type
    */
   MediaType getMediaType() const;
   /**
    * Set the media type to mt. This information is used for commands which require this parameter.
    *
    * @param mt The media type to use (HDDVD or BLURAY)
    */
   void setMediaType(MediaType mt);

   // These are the MMC commands, some may be composed of multiple commands actually
   /**
    * Reads the firmware string from the drive.
    *
    * @param drive_info The firmware string gets tored here. MUST be at least 65 bytes big
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int read_drive_info(/*out*/ char *drive_info);
   /**
    * Returns the data from the AACS Feature Descriptor.
    *
    * If a parameter is NULL this information won't get returned.
    * If the drive does not support this feature all returned values are 0 or false,
    * however it should be enough to check that version is 0.
    *
    * @param version AACS version
    * @param concurrent_agids Number of AGIDS that can be concurrently used
    * @param bng If true, the drive can generate Binding Nonces
    * @param block_count Number of blocks required to read / store Binding nonces
    * @param enabled If true, the AACS Feature is currently enabled
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int read_aacs_feature(/*out*/ unsigned char *version, /*out*/ unsigned char *concurrent_agids, /*out*/ bool *bng, /*out*/ unsigned char *block_count, /*out*/ bool *enabled);
   /**
    * Acquires an AGID.
    *
    * @param agid The AGID gets stored here. MUST be at least 1 byte big.
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int report_agid(/*out*/ unsigned char *agid);
   /**
    * Invalidates an AGID.
    *
    * @param agid The AGID to invalidate
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int invalidate_agid(unsigned char agid);
   /**
    * Sends a HOST CERTIFICATE CHALLENGE.
    *
    * @param agid The AGID to use
    * @param nonce The host nonce, MUST be 20 bytes.
    * @param cert The host certificate, MUST be 92 bytes.
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int send_host_cert_chal(unsigned char agid, const unsigned char *nonce, const unsigned char *cert);
   /**
    * Sends a HOST KEY.
    *
    * @param agid The AGID to use
    * @param point The point on the elliptic curve, MUST be 40 bytes.
    * @param signature The signature, MUST be 40 bytes.
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int send_host_key(unsigned char agid, const unsigned char *point, const unsigned char *signature);
   /**
    * Returns a DRIVE CERTIFICATE.
    *
    * @param agid The AGID to use
    * @param nonce The drive nonce gets stored here, MUST be 20 bytes.
    * @param cert The drive certificate gets stored here, MUST be 92 bytes.
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int report_drive_cert_chal(unsigned char agid, unsigned char *nonce, unsigned char *cert);
   /**
    * Returns a DRIVE KEY.
    *
    * @param agid The AGID to use
    * @param point The point on the elliptic curve gets stored here, MUST be 40 bytes.
    * @param signature The signature gets stored here, MUST be 40 bytes.
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int report_drive_key(unsigned char agid, unsigned char *point, unsigned char *signature);
   /**
    * Returns a VOLUME ID.
    *
    * @param agid The AGID to use
    * @param vid The volume id gets stored here, MUST be 16 bytes.
    * @param mac The volume id mac gets tored here, MUST be 16 bytes.
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int read_vid(unsigned char agid, /*out*/ unsigned char *vid, /*out*/ unsigned char *mac);
   /**
    * Returns a VOLUME ID using the XBox-Hack.
    *
    * @param agid The AGID to use. Can be any value in valid AGID space.
    * @param vid The volume id gets stored here, MUST be 16 bytes.
    * @param mac The volume id mac gets stored here, ;UST be 16 bytes. TODO: Is the MAC correct or just junk?
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int read_vid_mc08(unsigned char agid, /*out*/ unsigned char *vid, /*out*/ unsigned char *mac);
   /**
    * Returns a BINDING NONCE.
    *
    * @param agid The AGID to use
    * @param lba Starting LBA to read the Binding Nonce from
    * @param block_count Block Count for reading the Binding Nonce
    * @param bn The Binding Nonce gets stored here, MUST be 16 bytes
    * @param max The Binding Nonce Mac gets stored here, MUST be 16 bytes
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int read_bn(unsigned char agid, unsigned int lba, unsigned char block_count, /*out*/ unsigned char *bn, /*out*/ unsigned char *mac);

private:
   /**
    * Returns a KEY of the given format.
    *
    * @param agid The AGID to use
    * @param address Reserved / Address
    * @param block_count Reserved / Block Count
    * @param key_format KEY FORMAT, the KEY to return
    * @param buffer The returned key gets stored here, must be big enough for the requested key. Returns the raw data specified by key format.
    * @param length Size of buffer
    */
   int report_key(unsigned char agid, unsigned int address, unsigned char block_count, unsigned char key_format, /*out*/ unsigned char *buffer, unsigned short length);
   /**
    * Sends a KEY of the given format.
    *
    * @param agid The AGID to use.
    * @param key_format KEY FORMAT, the KEY to send
    * @param buffer The key to send, must be in the raw format specified by key format.
    * @param length Size of buffer
    * @return @see Drive::send_cmd(const unsigned char *, unsigned char *, size_t, size_t recv)
    */
   int send_key(unsigned char agid, unsigned char key_format, unsigned char *buffer, unsigned short length);
};


#endif
